@@ -17,7 +17,7 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 
-import sys, os
+import sys, os, re
 from owlready2 import *
 
 """
@@ -45,6 +45,10 @@ MINUS
 from owlready2.sparql.parser import *
 from owlready2.sparql.func   import register_python_function, FuncSupport
 
+_RE_AUTOMATIC_INDEX = re.compile(r"TABLE (.*?) AS (.*?) USING AUTOMATIC.*\((.*?)\)")
+_RE_NORMAL_INDEX    = re.compile(r"TABLE (.*?) AS (.*?) USING (COVERING )?INDEX (.*?) ")
+
+_DEPRIORIZE_SUBQUERIES_OPT = True
 
 class Translator(object):
   def __init__(self, world, error_on_undefined_entities = True):
@@ -103,7 +107,9 @@ class Translator(object):
             parameter_datatypes.append(number - 1)
         return "?%s" % r
       sql = re.sub("%s[^ ]*" % self.escape_mark, sub, sql)
-      
+
+    #sql = self.optimize_sql(sql)
+    
     if   self.main_query.type == "select":
       return PreparedSelectQuery(self.world, sql, [column.var for column in self.main_query.columns if not column.name.endswith("d")], [column.type for column in self.main_query.columns], nb_parameter, parameter_datatypes)
     
@@ -111,6 +117,91 @@ class Translator(object):
       select_param_indexes = [i - 1 for i in self.main_query.select_param_indexes]
       return PreparedModifyQuery(self.world, sql, [column.var for column in self.main_query.columns if not column.name.endswith("d")], [column.type for column in self.main_query.columns], nb_parameter, parameter_datatypes, self.world.get_ontology(self.main_query.ontology_iri.value) if self.main_query.ontology_iri else None, self.parse_inserts_deletes(self.main_query.deletes, self.main_query.columns), self.parse_inserts_deletes(self.main_query.inserts, self.main_query.columns), select_param_indexes)
     
+    
+  def optimize_sql(self, sql):
+    # Avoid Sqlite3 AUTOMATIC INDEX when a similar index can be used.
+    plan = list(self.world.graph.execute("""EXPLAIN QUERY PLAN %s""" % sql))
+    #for l in plan:
+      #if (" USING AUTOMATIC " in l[3]) and not (" TABLE " in l[3]): break
+    #else:
+    try:
+      self.world.graph.execute("PRAGMA automatic_index = false")
+      plan = list(self.world.graph.execute("""EXPLAIN QUERY PLAN %s""" % sql))
+    finally:
+      self.world.graph.execute("PRAGMA automatic_index = true")
+      
+    for l in plan:
+        match = _RE_NORMAL_INDEX.search(l[3])
+        if match:
+          table_type = match.group(1)
+          table_name = match.group(2)
+          index_name = match.group(4)
+          if table_name == "q": continue # Recursive hard-coded preliminary queries
+          #if ("%s INDEXED" % table_name) in sql: continue
+          table_def = "%s %s" % (table_type, table_name)
+          #print("OPTIMIZE!!!", l, table_type, table_name, index_name)
+          sql = sql.replace(table_def, "%s INDEXED BY %s" % (table_def, index_name), 1)
+    return sql
+    
+    
+    # while True:
+    #   changed = False
+    #   plan = list(self.world.graph.execute("""EXPLAIN QUERY PLAN %s""" % sql))
+      
+    #   for l in plan:
+    #     match = _RE_AUTOMATIC_INDEX.search(l[3])
+    #     if match:
+    #       table_type = match.group(1)
+    #       table_name = match.group(2)
+    #       #if table_name in indexeds: continue
+    #       cols       = { i[0] for i in match.group(3).split(" AND ") }
+          
+    #       if cols == { "s", "p", "o" }:
+    #         #print("OPTIMIZE!!!", l, table_type, table_name, cols)
+    #         table_def = "%s %s" % (table_type, table_name)
+    #         sql_s = sql.replace(table_def, "%s INDEXED BY index_%s_sp" % (table_def, table_type), 1)
+    #         sql_o = sql.replace(table_def, "%s INDEXED BY index_%s_op" % (table_def, table_type), 1)
+    #         sql_s = self.optimize_sql(sql_s)
+    #         sql_o = self.optimize_sql(sql_o)
+    #         s_nb_not_indexed = sql_s.count("NOT INDEXED")
+    #         o_nb_not_indexed = sql_o.count("NOT INDEXED")
+    #         if s_nb_not_indexed <= o_nb_not_indexed: sql = sql_s #; print("CHOIX s", s_nb_not_indexed, o_nb_not_indexed)
+    #         else:                                    sql = sql_o #; print("CHOIX o", s_nb_not_indexed, o_nb_not_indexed)
+    #         #indexeds.add(table_name)
+    #         changed = True
+    #         break
+          
+    #   if changed: continue
+            
+    #   for l in plan:
+    #     match = _RE_AUTOMATIC_INDEX.search(l[3])
+    #     if match:
+    #       table_type = match.group(1)
+    #       table_name = match.group(2)
+    #       cols       = { i[0] for i in match.group(3).split(" AND ") }
+    #       #print("OPTIMIZE!!!", l, table_type, table_name, cols)
+          
+    #       table_def = "%s %s" % (table_type, table_name)
+    #       if   "s" in cols:
+    #         sql = sql.replace(table_def, "%s INDEXED BY index_%s_sp" % (table_def, table_type), 1)
+    #         #indexeds.add(table_name)
+    #         changed = True
+    #       elif "o" in cols:
+    #         sql = sql.replace(table_def, "%s INDEXED BY index_%s_op" % (table_def, table_type), 1)
+    #         #indexeds.add(table_name)
+    #         changed = True
+    #       elif "p" in cols:
+    #         sql = sql.replace(table_def, "%s NOT INDEXED" % table_def, 1)
+    #         #indexeds.add(table_name)
+    #         changed = True
+            
+    #   if not changed: break
+    # return sql
+
+
+
+
+  
   def parse_inserts_deletes(self, triples, columns):
     var_2_column = { column.var : column for column in self.main_query.columns if not column.name.endswith("d") }
     r = []
@@ -148,11 +239,14 @@ class Translator(object):
       r.append(triple)
     return r
   
-  def new_sql_query(self, name, block, selects = None, distinct = None, solution_modifier = None, preliminary = False, extra_binds = None, nested_inside = None):
+  def new_sql_query(self, name, block, selects = None, distinct = None, solution_modifier = None, preliminary = False, extra_binds = None, nested_inside = None, copy_vars = False):
     if preliminary and not name: name = "prelim%s" % (len(self.preliminary_selects) + 1)
-
-    if isinstance(block, UnionBlock) and block.simple_union_triples: block = SimpleTripleBlock(block.simple_union_triples)
     
+    #print("BLOCK", name, block)
+    
+    if isinstance(block, UnionBlock) and block.simple_union_triples:
+      block = SimpleTripleBlock(block.simple_union_triples)
+      
     if   isinstance(block, SimpleTripleBlock):
       s = SQLQuery(name)
       
@@ -161,19 +255,20 @@ class Translator(object):
       s.optional = True
       
     elif isinstance(block, UnionBlock):
-      s = SQLCompoundQuery(name)
+      s = SQLCompoundQuery(name, nested_inside)
       if selects is None: selects = block.get_ordered_vars()
       
     elif isinstance(block, FilterBlock):
       s = SQLNestedQuery(name)
-      if isinstance(block, ExistsBlock): s.extra_sql = "IS NOT NULL"
-      else:                              s.extra_sql = "IS NULL"
+      #if isinstance(block, ExistsBlock): s.extra_sql = "IS NOT NULL"
+      #else:                              s.extra_sql = "IS NULL"
+      s.exists = isinstance(block, ExistsBlock)
       s.next_table_id = nested_inside.next_table_id
       s.vars = nested_inside.vars
       preliminary = False
       
     elif isinstance(block, NotExistsBlock):
-      s = SQLCompoundQuery(name)
+      s = SQLCompoundQuery(name, nested_inside)
       
     elif isinstance(block, SubQueryBlock):
       s = block.parse()
@@ -186,6 +281,14 @@ class Translator(object):
       s.preliminary = True
       self.preliminary_selects.append(s)
       
+      
+    # if copy_vars:
+    #   if hasattr(nested_inside, "vars"):
+    #     s.vars = nested_inside.vars.copy()
+    #   elif hasattr(nested_inside, "parent") and nested_inside.parent:
+    #     s.vars = nested_inside.parent.vars.copy()
+
+        
     extra_binds = extra_binds or []
     if isinstance(selects, list): # Otherwise, it is "SELECT *"
       for i, select in enumerate(selects):
@@ -211,7 +314,8 @@ class Translator(object):
       
     elif isinstance(block, UnionBlock):
       for alternative in block:
-        query = self.new_sql_query(None, alternative, selects, distinct, None, False, extra_binds)
+        #print("   ", s.parent.vars)
+        query = self.new_sql_query(None, alternative, selects, distinct, None, False, extra_binds, nested_inside = s, copy_vars = False)
         s.append(query, "UNION")
       s.finalize_compounds()
       
@@ -278,13 +382,14 @@ class PreparedQuery(object):
   
 class PreparedSelectQuery(PreparedQuery):
   def execute(self, params = ()):
+
     for l in PreparedQuery.execute(self, params):
       l2 = []
       i = 0
       while i < len(l):
         if self.column_types[i] == "objs":
           if l[i] is None: l2.append(None)
-          else:            l2.append(self.world._to_python(l[i], None) or l[i])
+          else: l2.append(self.world._to_python(l[i], None) or l[i])
           i += 1
         else:
           if l[i + 1] is None:
@@ -295,6 +400,110 @@ class PreparedSelectQuery(PreparedQuery):
           i += 2
       yield l2
       
+  def execute_csv(self, params = (), separator = ","):
+    import csv, io
+    b = io.StringIO()
+    f = csv.writer(b, delimiter = separator)
+    f.writerow(col[1:] for col in self.column_names)
+    rows = []
+
+    for l in PreparedQuery.execute(self, params):
+      l2 = []
+      i = 0
+      while i < len(l):
+        if self.column_types[i] == "objs":
+          if   l[i] is None: l2.append("")
+          elif l[i] > 0:     l2.append(self.world._unabbreviate(l[i]))
+          else:              l2.append("_:%s" % (-l[i]))
+          i += 1
+        else:
+          if l[i + 1] is None:
+            if   l[i] is None: l2.append("")
+            elif l[i] > 0:     l2.append(self.world._unabbreviate(l[i]))
+            else:              l2.append("_:%s" % (-l[i]))
+          else:
+            l2.append(str(self.world._to_python(l[i], l[i + 1])))
+          i += 2
+      f.writerow(l2)
+    return b.getvalue()
+  
+  def execute_tsv(self, params = ()): return self.execute_csv(params, "\t")
+
+  def execute_json(self, params = ()):
+    bindings = []
+    colnames = [col[1:] for col in self.column_names]
+    json = { "head" : { "vars" : colnames },
+             "results" : { "bindings" : bindings } }
+    for l in PreparedQuery.execute(self, params):
+      binding = {}
+      bindings.append(binding)
+      i = 0
+      c = 0
+      while i < len(l):
+        if self.column_types[i] == "objs":
+          if   l[i] is None: pass
+          elif l[i] > 0:     binding[colnames[c]] = { "type" : "uri", "value" : self.world._unabbreviate(l[i]) }
+          else:              binding[colnames[c]] = { "type" : "bnode", "value" : "r%s" % (-l[i]) }
+          i += 1
+        else:
+          if l[i + 1] is None:
+            if   l[i] is None: pass
+            elif l[i] > 0:     binding[colnames[c]] = { "type" : "uri", "value" : self.world._unabbreviate(l[i]) }
+            else:              binding[colnames[c]] = { "type" : "bnode", "value" : "r%s" % (-l[i]) }
+          else:
+            value = str(self.world._to_python(l[i], l[i + 1]))
+            if   isinstance(l[i + 1], str): binding[colnames[c]] = { "type" : "literal", "value" : value, "xml:lang" : l[i + 1][1:] }
+            elif l[i + 1]:                  binding[colnames[c]] = { "type" : "literal", "value" : value, "datatype" : self.world._unabbreviate(l[i + 1]) }
+            else:                           binding[colnames[c]] = { "type" : "literal", "value" : value }
+          i += 2
+        c += 1
+    return repr(json)
+
+  def execute_xml(self, params = ()):
+    bindings = []
+    colnames = [col[1:] for col in self.column_names]
+    xml = """<?xml version="1.0"?>
+<sparql xmlns="http://www.w3.org/2005/sparql-results#">
+  <head>
+"""
+    for colname in colnames:
+      xml += """    <variable name="%s"/>\n""" % colname
+    xml += """  </head>
+  <results>
+"""
+    
+    for l in PreparedQuery.execute(self, params):
+      xml += """    <result>\n"""
+      i = 0
+      c = 0
+      while i < len(l):
+        xml += """      <binding name="%s">\n""" % colnames[c]
+        if self.column_types[i] == "objs":
+          if   l[i] is None: pass
+          elif l[i] > 0:     xml += """        <uri>%s</uri>\n""" % self.world._unabbreviate(l[i])
+          else:              xml += """        <bnode>r%s</bnode>\n""" % (-l[i])
+          i += 1
+        else:
+          if l[i + 1] is None:
+            if   l[i] is None: pass
+            elif l[i] > 0:     xml += """        <uri>%s</uri>\n""" % self.world._unabbreviate(l[i])
+            else:              xml += """        <bnode>r%s</bnode>\n""" % (-l[i])
+          else:
+            value = str(self.world._to_python(l[i], l[i + 1]))
+            if   isinstance(l[i + 1], str): xml += """        <literal xml:lang="%s">%s</literal>\n""" % (l[i + 1][1:], value)
+            elif l[i + 1]:                  xml += """        <literal datatype="%s">%s</literal>\n""" % (self.world._unabbreviate(l[i + 1]), value)
+            else:                           xml += """        <literal>%s</literal>\n""" % value
+          i += 2
+        c += 1
+        xml += """      </binding>\n"""
+      xml += """    </result>\n"""
+        
+    xml += """  </results>
+</sparql>
+"""
+    return xml
+  
+  
 class PreparedModifyQuery(PreparedQuery):
   def __init__(self, world, sql, column_names, column_types, nb_parameter, parameter_datatypes, ontology, deletes, inserts, select_param_indexes):
     PreparedQuery.__init__(self, world, sql, column_names, column_types, nb_parameter, parameter_datatypes)
@@ -357,9 +566,24 @@ class Variable(object):
     self.bindings       = []
     self.bind           = None
     self.initial_query  = None
+    self.nb_table       = 0
     
   def __repr__(self): return """<Variable %s type %s, %s bindings>""" % (self.name, self.type, len(self.bindings))
   
+  def get_binding(self, query):
+    if not self.bindings:
+      #print("* Owlready2 * WARNING: variable without binding in SPARQL, use a suboptimal option", file = sys.stderr)
+      table = Table("any%s" % query.next_table_id, """(SELECT DISTINCT s AS o, NULL AS d FROM quads UNION SELECT DISTINCT o, d FROM quads)""")
+      table.query = query
+      query.tables.append(table)
+      query.name_2_table[table.name] = table
+      query.next_table_id += 1
+      self.bindings.append("%s.o" % table.name)
+    i = 0
+    for binding in self.bindings:
+      if not binding.startswith("IN "): break
+    return binding
+  
   def update_type(self, type):
     if   self.type == "quads": self.type = type
     elif (type != self.type) and (type != "quads") :
@@ -382,23 +606,24 @@ class Table(object):
   
 class SQLQuery(FuncSupport):
   def __init__(self, name):
-    self.name                    = name
-    self.preliminary             = False
-    self.recursive               = False
-    self.translator              = CURRENT_TRANSLATOR.get()
-    self.distinct                = False
-    self.raw_selects             = None
-    self.next_table_id           = 1
-    self.tables                  = []
-    self.name_2_table            = {}
-    self.columns                 = []
-    self.conditions              = []
-    self.triples                 = []
-    self.vars_needed_for_select  = set()
-    self.vars                    = {}
-    self.extra_sql               = ""
-    self.select_simple_union     = False
-    self.optional                = False
+    self.name                     = name
+    self.preliminary              = False
+    self.recursive                = False
+    self.translator               = CURRENT_TRANSLATOR.get()
+    self.distinct                 = False
+    self.raw_selects              = None
+    self.next_table_id            = 1
+    self.tables                   = []
+    self.name_2_table             = {}
+    self.columns                  = []
+    self.conditions               = []
+    self.triples                  = []
+    self.vars_needed_for_select   = set()
+    self.vars                     = {}
+    self.extra_sql                = ""
+    self.select_simple_union      = False
+    self.optional                 = False
+    self.has_one                  = False
     
   def __repr__(self): return "<%s '%s'>" % (self.__class__.__name__, self.sql())
   
@@ -442,8 +667,7 @@ class SQLQuery(FuncSupport):
     if isinstance(x, rply.Token): name = x.value
     else:                         name = x
     var = self.vars.get(name)
-    if not var:
-      self.vars[name] = var = Variable(name)
+    if not var: self.vars[name] = var = Variable(name)
     return var
   
   def parse_selects(self, selects):
@@ -479,9 +703,12 @@ class SQLQuery(FuncSupport):
     sql = self.parse_expression(filter.constraint)
     self.conditions.append(sql)
     
-  def add_subqueries(self, sub):
+  def add_subquery(self, sub):
     if isinstance(sub, SQLNestedQuery):
       self.conditions.append(sub)
+      if _DEPRIORIZE_SUBQUERIES_OPT and not self.has_one:
+        self.tables.append(Table("one", "one"))
+        self.has_one = True
     else:
       table = Table("p%s" % self.next_table_id, sub.name)
       table.query = sub
@@ -493,7 +720,7 @@ class SQLQuery(FuncSupport):
         conditions = table.join_conditions
       else:
         conditions = self.conditions
-        
+
       for column in sub.columns:
         var = self.parse_var(column.var)
         var.update_type(column.type)
@@ -552,16 +779,6 @@ class SQLQuery(FuncSupport):
         if   (s.name != "VAR"): fixed = "s"
         elif (o.name != "VAR"): fixed = "o"
         else:
-          # nb_s = nb_o = 0
-          # for triple2 in self.triples:
-          #   for x in triple2.var_names:
-          #     if x == s.value: nb_s += 1
-          #     if x == o.value: nb_o += 1
-          # if   nb_s == 1:    fixed = "o"
-          # elif nb_o == 1:    fixed = "s"
-          # elif nb_s <= nb_o: fixed = "s"
-          # else:              fixed = "o"
-          # print("NB", nb_s, nb_o)
           fix_levels = self.get_fix_levels([self.parse_var(s), self.parse_var(o)], triple)
           if fix_levels[self.parse_var(s)] >= fix_levels[self.parse_var(o)]: fixed = "s"
           else:                                                              fixed = "o"
@@ -590,8 +807,44 @@ class SQLQuery(FuncSupport):
         
     selected_non_preliminary_triples = frozenset(self.extract_triples(non_preliminary_triples, self.vars_needed_for_select))
     selected_non_preliminary_triples = selected_non_preliminary_triples | remnant_triples
-
-    for triple in self.triples: # Pass 4: create triples tables and conditions
+    
+    
+    vars_needing_binding = set(self.vars_needed_for_select)
+    for triple in self.triples: # Pass 4: compute number of bindings per variable
+      if   isinstance(triple, (Bind, Filter)):
+        for var_name in triple.var_names:
+          vars_needing_binding.add(self.parse_var(var_name))
+        continue
+      if   isinstance(triple, Block): continue
+      s, p, o = triple
+      if (not p.modifier) and (not triple in selected_non_preliminary_triples): continue
+      if triple.local_table_type.startswith("prelim") and not triple.consider_p: continue
+      for (x, consider) in zip(triple, [triple.consider_s, triple.consider_p, triple.consider_o]):
+        if consider and (x.name == "VAR"):
+          x = self.parse_var(x)
+          x.nb_table += 1
+          
+    for triple in self.triples: # Pass 5: pre-register bindings for preliminary table
+      if   isinstance(triple, (Bind, Filter, Block)): continue
+      s, p, o = triple
+      if (not p.modifier) and (not triple in selected_non_preliminary_triples): continue
+      
+      triple.to_skip = False
+      if self.raw_selects == "*": continue
+      if triple.local_table_type.startswith("prelim") and not triple.consider_p:
+        if (s.name == "VAR") and (self.parse_var(s) in vars_needing_binding) and self.parse_var(s).nb_table == 0: continue
+        if (o.name == "VAR") and (self.parse_var(o) in vars_needing_binding) and self.parse_var(o).nb_table == 0: continue
+        triple.to_skip = True
+        extra = ""
+        if p.modifier == "+": extra = " WHERE nb>0"
+        if s.name == "VAR":
+          s = self.parse_var(s)
+          s.bindings.insert(0, """IN (SELECT s FROM %s%s)""" % (triple.local_table_type, extra))
+        if o.name == "VAR":
+          o = self.parse_var(o)
+          o.bindings.insert(0, """IN (SELECT o FROM %s%s)""" % (triple.local_table_type, extra))
+          
+    for triple in self.triples: # Pass 6: create triples tables and conditions
       if   isinstance(triple, Bind):
         self.parse_bind(triple)
         continue
@@ -599,21 +852,23 @@ class SQLQuery(FuncSupport):
         self.parse_filter(triple)
         continue
       elif isinstance(triple, Block):
-        if   isinstance(triple, SimpleTripleBlock) and len(triple) == 0: # Empty
-          continue
+        if   isinstance(triple, SimpleTripleBlock) and (len(triple) == 0): continue # Empty
         elif isinstance(triple, FilterBlock):
           sub = self.translator.new_sql_query(None, triple, nested_inside = self)
-          self.add_subqueries(sub)
+          self.add_subquery(sub)
           continue
         else:
-          sub = self.translator.new_sql_query(None, triple, preliminary = True)
-          self.add_subqueries(sub)
+          sub = self.translator.new_sql_query(None, triple, preliminary = True, nested_inside = self, copy_vars = True)
+          self.add_subquery(sub)
           continue
-        
+      if triple.to_skip: continue
+      
       s, p, o = triple
       if (not p.modifier) and (not triple in selected_non_preliminary_triples): continue
       
-      table = Table("q%s" % self.next_table_id, triple.local_table_type)
+      if self.name == "main": select_name = ""
+      else:                   select_name = "%s_" % self.name
+      table = Table("%sq%s" % (select_name, self.next_table_id), triple.local_table_type)
       if triple.optional:
         table.join = "LEFT JOIN"
         conditions = table.join_conditions
@@ -623,21 +878,11 @@ class SQLQuery(FuncSupport):
       self.tables.append(table)
       self.name_2_table[table.name] = table
       
-      s_fixed = triple.consider_s and self.is_fixed(s)
-      o_fixed = triple.consider_o and self.is_fixed(o)
-      if   s_fixed and o_fixed: # Favor index SP over index OP, because there are usually more diversity in the values of S than in O
-        if   table.type == "objs":  table.index = "index_objs_sp"
-        elif table.type == "datas": table.index = "index_datas_sp"
-      elif s_fixed: # Favor index SP over SQLite3 partial covering index, because benchmark suggests better performances
-        if   table.type == "objs":  table.index = "index_objs_sp"
-        elif table.type == "datas": table.index = "index_datas_sp"
-        
       if triple.consider_s: self.create_conditions(conditions, table, "s", s)
-      if triple.consider_p: self.create_conditions(conditions, table, "p", p)
-      if triple.consider_o: self.create_conditions(conditions, table, "o", o)
+      if triple.consider_p: self.create_conditions(conditions, table, "p", p, triple.likelihood_p)
+      if triple.consider_o: self.create_conditions(conditions, table, "o", o, triple.likelihood_o)
       
       if p.modifier == "+": conditions.append("%s.nb>0"  % table.name)
-      
 
   def get_fix_levels(self, vars0, exclude_triple = None):
     vars0_names = { var.name for var in vars0 }
@@ -692,14 +937,17 @@ class SQLQuery(FuncSupport):
     return fix_levels
       
   def extract_triples(self, triples, vars, except_triple = None):
+    #print("   EXTRACT", triples, vars, except_triple)
     var_names = { var.name for var in vars }
     while True:
       r = [triple for triple in triples if (not triple == except_triple) and isinstance(triple, (Triple, Filter, Bind)) and (not triple.var_names.isdisjoint(var_names))]
       var_names2 = { var_name for triple in r for var_name in triple.var_names }
-      if var_names2 == var_names: return r
+      if var_names2 == var_names:
+        #print("   EXTRACTED", r)
+        return r
       var_names = var_names2
       
-  def create_conditions(self, conditions, table, n, x):
+  def create_conditions(self, conditions, table, n, x, likelihood = None):
     if isinstance(x, SpecialCondition):
       x.create_conditions(conditions, table, n)
     else:
@@ -713,14 +961,22 @@ class SQLQuery(FuncSupport):
               conditions.append("%s.%sd IS NOT NULL" % (table.name, n[:-1])) # Datatype part
             break
           
-      if not sql   is None: conditions.append("%s.%s=%s"  % (table.name, n,      sql))
-      if not sql_d is None: conditions.append("%s.%sd=%s" % (table.name, n[:-1], sql_d)) # Datatype part
-      
+      if not sql is None:
+        if isinstance(sql, str) and sql.startswith("IN "): operator = " "; sqls = self.parse_var(x).bindings
+        else:                                              operator = "="; sqls = [sql]
+        
+        for sql in sqls:
+          condition = "%s.%s%s%s"  % (table.name, n, operator, sql)
+          if likelihood is None: conditions.append(condition)
+          else:                  conditions.append("LIKELIHOOD(%s, %s)" % (condition, likelihood))
+      if not sql_d is None:
+        conditions.append("%s.%sd=%s" % (table.name, n[:-1], sql_d)) # Datatype part
+        
       if x.name == "VAR": x = self.vars[x.value]
       if   isinstance(x, Variable):
-        if not x.bindings: x.initial_query = self
+        if not x.initial_query: x.initial_query = self
         x.bindings.append("%s.%s" % (table.name, n))
-
+        
   def is_fixed(self, x):
     if x.name != "VAR": return True
     x = self.parse_var(x)
@@ -731,7 +987,7 @@ class SQLQuery(FuncSupport):
   def finalize_columns(self):
     selected_parameter_index = 0
     i = j = 0
-    if self.raw_selects == "*": selects = [self.vars[var] for var in self.block.get_ordered_vars() if not var.startswith("??")]
+    if self.raw_selects == "*": selects = [self.vars[var] for var in self.block.get_ordered_vars() if not var.startswith("??")]      
     else:                       selects = self.raw_selects
     
     def do_select(select):
@@ -762,17 +1018,22 @@ class SQLQuery(FuncSupport):
       else:
         var_name, sql, sql_type, sql_d, sql_d_type = do_select(select)
         
-      if sql is None: raise ValueError("Cannot select '%s'!" % select)
+      if sql is None:
+        if not self.name: # Inside a UNION => the variable is not available in this alternative of th UNION
+          sql = "NULL"
+          sql_type = "objs"
+        else:
+          raise ValueError("Cannot select '%s'!" % select)
       self.columns.append(Column(var_name, sql_type,   sql,   "col%s_o" % i, j)); j += 1
       if not sql_d is None: self.columns.append(Column(var_name, sql_d_type, sql_d, "col%s_d" % i, j)); j += 1
-        
+      
   def _to_sql(self, x):
     if isinstance(x, rply.Token) and (x.name == "VAR"): x = self.parse_var(x)
 
     if   isinstance(x, str): return x, "value", None, None
     elif isinstance(x, Variable):
       if not x.bindings: return None, None, None, None
-      binding = x.bindings[0]
+      binding = x.get_binding(self)
       
       if   x.type == "objs":  return binding, "objs", None, None
       else:
@@ -794,15 +1055,16 @@ class SQLQuery(FuncSupport):
     
 class SQLCompoundQuery(object):
   recursive = False
-  def __init__(self, name):
+  def __init__(self, name, parent):
     self.name                    = name
+    self.parent                  = parent
     self.translator              = CURRENT_TRANSLATOR.get()
     self.queries                 = []
     self.preliminary             = False
     self.optional                = False
     
   def __repr__(self): return "<%s '%s'>" % (self.__class__.__name__, self.sql())
-
+  
   def append(self, query, operator = ""):
     query.operator = operator
     self.queries.append(query)
@@ -836,18 +1098,31 @@ class SQLCompoundQuery(object):
 class SQLNestedQuery(SQLQuery):
   def __init__(self, name):
     SQLQuery.__init__(self, name)
+    self.exists = True
     
   def finalize_columns(self):
     self.columns = [Column(None, "datas", "1", "col1_o", 1)]
     
   def sql(self):
-    extra_sql = self.extra_sql
-    self.extra_sql = ""
+    #extra_sql = self.extra_sql
+    #self.extra_sql = ""
     
-    sql = "(%s)" % SQLQuery.sql(self)
+    #sql = "(%s)" % SQLQuery.sql(self)
     
-    if extra_sql: sql += " %s" % extra_sql
-    self.extra_sql = extra_sql
+    #if extra_sql: sql += " %s" % extra_sql
+    #self.extra_sql = extra_sql
+    
+    sql = SQLQuery.sql(self)
+
+    if _DEPRIORIZE_SUBQUERIES_OPT:
+      if   self.exists == True:  sql =     "one.i= EXISTS(%s)" % sql
+      elif self.exists == False: sql = "NOT one.i= EXISTS(%s)" % sql
+      else:                      sql = "(%s)" % sql
+    else:
+      if   self.exists == True:  sql =     "EXISTS(%s)" % sql
+      elif self.exists == False: sql = "NOT EXISTS(%s)" % sql
+      else:                      sql = "(%s)" % sql
+      
     return sql
   
   def __str__(self): return self.sql()
